/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  Because we have our own SDI handler, the interrupt does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */

#include "module.h"
#define UART_UI
#undef UART_FIFO /* not enough space */


#if 0
#define LOUDNESS_ALWAYS_ON_OFF
#endif


/*
  - subdirectory support

  NEW features for SCI_UI:
  - play mode:
    + normal continuous play
    + pause before song (doubles as pause after reset)
    + loop song
    + report file found / not found and wait for play command
  - find file by number but have the file name readable from memory

  TODO features:

  Extra features
  - L/R VU-meter through SCI
 */


#include <string.h>
#ifdef UART_FIFO
#include <h1103/parametric.h>
#include "uartctrl.h"
#endif

auto void play_frame(void);
//void Disable();
//void Enable();
//void Reset();
void putch(register __a0 int ch);
void putch15(register __a0 int ch);
void putword(register __a0 int ch);

//#undef putchar
//#undef puts

//#include <stdio.h>

#include <stdlib.h>

#ifdef WAIT_FOR_UART_BYTE
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC))
#else
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif/*elseWAIT_FOR_UART_BYTE*/


//#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */

#ifndef NOCHECKFILETYPE
#define CHECKFILETYPE /* vs1011 29 words? check file type according to short-filename suffix */
#endif/*NOCHECKFILETYPE*/

#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */

void putch(register __a0 int ch);
void putch15(register __a0 int ch);

#ifdef CHECKFILETYPE
#undef CHECKFILETYPE /*temporary undef*/
#define CHECKFILETYPE CheckFileType
#endif

#define CTRL3_UPDATE_VOL        (1<<15)
#define CTRL3_I2S_ENABLE        (1<<9) /*VS1053: enable I2S*/
#define CTRL3_BY_NAME           (1<<8) /*locate by name*/
#define CTRL3_RECORD_ON         (1<<7)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

//#define SPI_WREN  0x06
//#define SPI_WRDI  0x04
//#define SPI_RDSR  0x05
//#define SPI_WRSR  0x01
#define SPI_READ  0x03
//#define SPI_WRITE 0x02

typedef enum {
    spi_pause=0,
    spi_file,
    spi_seek,
    spi_waitdata,
    spi_data,
    spi_zeros,
    spi_record,
    spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
s_int16 spiCnt;

struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;    /*0x1811,0x1812*/
    u_int32 sector;  /*0x1813,0x1814*/
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
  u_int16 err; //for ReadDiskSector()
} spix;


#define 	MMC_GO_IDLE_STATE   0
#define 	MMC_SEND_OP_COND   1
#define 	MMC_SEND_IF_COND   8
#define 	MMC_SEND_CSD   9
#define 	MMC_SEND_CID   10
#define 	MMC_SEND_STATUS   13
#define 	MMC_SET_BLOCKLEN   16
#define 	MMC_READ_SINGLE_BLOCK   17
#define 	MMC_WRITE_BLOCK   24
#define 	MMC_PROGRAM_CSD   27
#define 	MMC_SET_WRITE_PROT   28
#define 	MMC_CLR_WRITE_PROT   29
#define 	MMC_SEND_WRITE_PROT   30
#define 	MMC_TAG_SECTOR_START   32
#define 	MMC_TAG_SECTOR_END   33
#define 	MMC_UNTAG_SECTOR   34
#define 	MMC_TAG_ERASE_GROUP_START   35
#define 	MMC_TAG_ERARE_GROUP_END   36
#define 	MMC_UNTAG_ERASE_GROUP   37
#define 	MMC_ERASE   38
#define         MMC_READ_OCR     58
#define 	MMC_CRC_ON_OFF   59
#define 	MMC_R1_BUSY   0x80
#define 	MMC_R1_PARAMETER   0x40
#define 	MMC_R1_ADDRESS   0x20
#define 	MMC_R1_ERASE_SEQ   0x10
#define 	MMC_R1_COM_CRC   0x08
#define 	MMC_R1_ILLEGAL_COM   0x04
#define 	MMC_R1_ERASE_RESET   0x02
#define 	MMC_R1_IDLE_STATE   0x01
#define 	MMC_STARTBLOCK_READ   0xFE
#define 	MMC_STARTBLOCK_WRITE   0xFE
#define 	MMC_STARTBLOCK_MWRITE   0xFC
#define 	MMC_STOPTRAN_WRITE   0xFD
#define 	MMC_DE_MASK   0x1F
#define 	MMC_DE_ERROR   0x01
#define 	MMC_DE_CC_ERROR   0x02
#define 	MMC_DE_ECC_FAIL   0x04
#define 	MMC_DE_OUT_OF_RANGE   0x04
#define 	MMC_DE_CARD_LOCKED   0x04
#define 	MMC_DR_MASK   0x1F
#define 	MMC_DR_ACCEPT   0x05
#define 	MMC_DR_REJECT_CRC   0x0B
#define 	MMC_DR_REJECT_WRITE_ERROR   0x0D

#if 1
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
    register __b1 u_int16 t;
#if 1
    puts("Cmd(");
    puthex(cmd);
    puthex(arg>>16);
    puthex(arg);
    puts(")");
#endif
    /* some MMC brands require the extra clock pulses     */
    SpiSendClocks();
    SpiSendClocks();
    USEX(GPIO_ODATA) &= ~SPI_xCS2;

    while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
	/* Unexpected busy signal from MMC */
#if 0
	putchar('u');
	putchar('\n');
#endif
    }
    SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
    SpiSendReceiveMmc(arg>>16, 16);
    SpiSendReceiveMmc(arg, 16);
    t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */

    cmd = 100;               /* timeout value, try max. 100 times */
    while (cmd && (t & 0x80)) { /* R1 response not detected */
	t = SpiSendReceiveMmc(0xff00, 8);
	cmd--;               /* decrease the timeout value */
    }
#if 0
    puthex(t);
    putchar('\n');
#endif
    return t;
}
#endif

#if 1
auto void ReadDiskSector(register __a u_int32 sector);
#else
auto void ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;

    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
    do {
	i = SpiSendReceiveMmc(0xff00,8);
    } while (i == 0xff);

#if 0
    if (i != 0xfe) {
	SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
	return;
    }
#endif

    for (i=0; i<512; i+=2) {
	*buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return;
}
#endif

#if 0
struct Fat {
    byte BS_jmpBoot[3];		/**< 0 x86 Boot Jump Code */
    byte BS_OEMName[8];		/**< 3 Formatter's name, usually "MSWIN4.1" */
    
    word BPB_BytsPerSec;	/**< 11 Bytes per sector (512) */
    byte BPB_SecPerClus;	/**< 13 Sectors per Cluster (1,2,4,8,..,128) */
    word BPB_RsvdSecCnt;	/**< 14 Reserved sectors (1 (32 for FAT32)) */
    byte BPB_NumFATs;		/**< 16 Number of FATs (2) */
    word BPB_RootEntCnt;	/**< 17 FAT12/16 n of root dir entries */
    word BPB_TotSec16;		/**< 19 Old sector count (0 for FAT32) */
    byte BPB_Media;		/**< 21 Media Type (eg 0xF8) */
    word BPB_FATSz16;		/**< 22 Size of one FAT16 in sectors */
    word BPB_SecPerTrk;		/**< 24 Old CHS Sectors Per Track */
    word BPB_NumHeads;		/**< 26 Old CSH Number of Heads */
    u_32 BPB_HiddSec;		/**< 28 n of sectors before this volume */
    u_32 BPB_TotSec32;		/**< 32 New sector count (0 for FAT12/16) */
   
    /** FAT type specific extensions */
    union Extensions{	

      /** FAT12/16 specific extensions to Bios Parameter Block*/
      struct Fat16Specific {	
	byte BS_DrvNum;		/**< 36 DOS INT13 Drive Number (0x80=HD)*/
	byte BS_Reserved1;	/**< 37 For WINNT; Format to 0 */
	byte BS_BootSig;	/**< 38 0x29 if next 3 fields are present */
	byte BS_VolID[4];	/**< 39 Volume ID (usually format datetime) */
	byte BS_VolLab[11];	/**< 43 Volume Label */
	byte BS_FilSysType[8];	/**< 54 Decorative name of fs, eg "FAT16   "*/
      } _16;

      /** FAT32 specific extensions to Bios Parameter Block*/
      struct Fat32Specific {	
        u_32 BPB_FATSz32;	/**< 36 Size of one FAT32 in sectors */
	word BPB_ExtFlags;	/**< 40 Flags; active FAT number etc */
	word BPB_FSVer;		/**< 42 File System Version (0x0000) */
	u_32 BPB_RootClus;	/**< 44 Start cluster of Root Dir. (2) */
	word BPB_FSInfo;	/**< 48 Start sector of FSINFO in Resvd area */
	word BPB_BkBootSec;	/**< 50 Sector in Resvd area for BkBoot (6) */
	byte BPB_Reserved[12];	/**< 52 Reserved, Always 0. */
	byte BS_DrvNum;      	/**< 64 DOS INT13 Drive Number (0x80=HD) */
	byte BS_Reserved1;      /**< 65 For WINNT, Format to 0 */
	byte BS_BootSig;	/**< 66 0x29 if next 3 fields are present */
	byte BS_VolID[4];	/**< 67 Volume ID (usually format datetime) */
	byte BS_VolLab[11];	/**< 71 Volume Label */
	byte BS_FilSysType[8];  /**< 82 Decorative name of fs, eg "FAT32   "*/
      } _32;
    } ext;
  } fat;
#endif

#if 1
auto u_int16 GetByte(register __c0 u_int16 n);
#else
auto u_int16 GetByte(register __c0 u_int16 n) {
    register u_int16 *p = stream_buffer[n>>1];
    if (n & 1) {
	n = *p;
    } else {
	n = *p >> 8;
    }
    return n & 0xff;
}
#endif
#if 1
auto u_int16 GetWord(register __c0 u_int16 n);
#else
auto u_int16 GetWord(register __c0 u_int16 n) {
    return GetByte(n) | (GetByte(n+1)<<8);
}
#endif
#if 1
auto u_int32 GetLong(register __c0 u_int16 n);
#else
auto u_int32 GetLong(register __c0 u_int16 n) {
    return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
}
#endif

extern __y u_int16 fileName[6];
#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif

#if 1
auto u_int16 InitFileSystem(void);
#else
auto u_int16 InitFileSystem(void) {
    /* Load MBR -- the first sector on disk */
    ReadDiskSector(0);

    /* Ok, it should be a MBR sector. Let's verify */
    if (stream_buffer[255] != 0x55aa) {
#ifdef USE_DEBUG
//      puthex(stream_buffer[255]);
	puts("!MBR");
#endif
	return -1; /* sector 0 is not MBR. */
    }

#if 0
    /* This checks that partition 1 is active. Alter code to allow
       other partition configurations. */
    if (GetByte(0x1be) == 0x80) {
    } else {
	puthex(GetByte(0x1be));
	puts(" PF!");
    }
#endif

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
	GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
	) {
	fatInfo.currentSector = GetLong(0x1c6);
	ReadDiskSector(fatInfo.currentSector);
    }
    {
	u_int16 BPB_BlksPerSec;

	/* Determine FAT Type (16/32) */
	/* This should be done better, but it'll do for now. */
	fatInfo.BPB_RootEntCnt = GetWord(17);
	/* FAT32 does not have separate root entries. */
	/* TODO: should detect FAT32 by BPB_FATSz16==0 */
	fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
	fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */
#if 0
	puthex(fatInfo.FilSysType);
	puts("fsys");
#endif
#if 0
	if (!IS_FAT_32){
	    if (GetByte(54+4)=='2') {/*diskSect.fat.ext._16.BS_FilSysType[4]=='2'*/) {
		fatInfo.IS_FAT_12 = GetByte(54);
	    }
	}
#endif

#if 0
	if (fatInfo.IS_FAT_32) {
	    puts("FAT32");
	} else {
	    puthex(fatInfo.BPB_RootEntCnt);
	    puts(" FAT");
	    puthex(GetWord(19));
	    puts(" TSc16");
	}
	puthex(GetWord(54+3)); /* 3231 == '12' 3631= '16' 0000=FAT32 */
	puts(" fst");
#endif
#if 0
	putchar('\"');
	{
	    register int i;
	    for (i=0; i<11; i++) {
		if (fatInfo.IS_FAT_32) {
		    putchar(GetByte(71+i));
		} else {
		    putchar(GetByte(43+i));
		}
	    }
	}
	putchar('\"');
	putchar('\n');
#endif
	/* OK, let's calculate */
	/* First, let's get rid of the idea that we have byte addresses
	   in the file system. Nope, let's only deal in physical disk
	   sectors of 512 bytes. First we convert the FAT byter per sector
	   value to "512B disk sectors per fat sector" value. */

	BPB_BlksPerSec = GetWord(11) >> 9;

	/* Then we adjust the Sector per Cluster to mean physical disk
	   sectors. in 99% of the cases it is already so because bytes
	   per sector almost always is 512 in FAT. Maximum cluster size
	   is 65536 bytes (128 disk sectors). */

	fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
#if 0
	puthex(fatInfo.fatSectorsPerCluster);
	puts(" FatSPC");
#endif
	fatInfo.fatStart = fatInfo.currentSector +
	    (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
#if 0
	puthex(fatInfo.fatStart);
	puts(" fatS");
#endif
	fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
	if (fatInfo.rootStart == 0) {
	    if (!fatInfo.IS_FAT_32) 
		return 0x0b; /* should be FAT32; can not find root directory */
	    fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
	}
	/* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
	fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
						 BPB_BlksPerSec)
	    + fatInfo.fatStart;
#if 0
	puthex(fatInfo.rootStart);
	puts(" rootS");
#endif
	fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
	    - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/

#if 0
	puthex(fatInfo.dataStart);
	puts(" dataS");
//	puthex(fatInfo.clusters);
//	puts(" Clustrs");
#endif
    }
    return 0;
}
#endif

#if 0
/** FAT/VFAT directory record union */
union DirRecordUnion {
  /** Standard directory entry */
  struct Entry {
    byte Name[11]; 0
    byte Attr;     11
    byte NTRes;    12
    byte CrtTimeTenth; 13
    word CrtTime;  14
    word CrtDate;  16
    word LstAccDate; 18
    word FstClusHi; 20
    word WrtTime;   22
    word WrtDate;   24
    word FstClusLo; 26
    u_32 FileSize;  28
  } entry;

  /** Extended directory entry */
  struct LongEntry {
    byte Ord;      0
    word Name1[5]; 1 /**< characters 1-5 */
    byte Attr;     11
    byte Type;     12 /**< entry type, zero=long name component */
    byte Chksum;   13
    word Name2[6]; 14 /**< characters 6-11 */
    word FstClusLO; 26 /**< zero for long entry */
    word Name3[2]; 28 /**< characters 12-13 */
  } longentry;
};
#endif


#define LAST_FRAGMENT 0x80000000UL
struct FRAGMENT {
    u_int32 start; /*high bit set if last fragment*/
    u_int16 size;
};
//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster);
#else
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster) {
    register __d0 u_int16 offset;

    /* fatCluster means now current cluster, starting from 0, not 2. */
    /* note that startcluster=2 is already compensated in fatInitGlobals */
#if 0
    puthex(fatCluster>>16);
    puthex(fatCluster);
    puts(" fatC");
#endif
#if 0
    frag->size = 0;
    frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
#endif
    goto l55; /* read first sector */

    while (1) {
	u_int32 t;
	if (fatInfo.IS_FAT_32) {
	    t = GetLong(offset) & 0x0fffffffUL;
	} else {
#if 0
	    {
		register int i;
		for (i=0;i<256;i++) {
		    t = GetByte(i);
		    puthex(t);
		    putchar(' ');
		    if ((i&7)==7)
			putchar('\n');
		}
	    }
#endif

	    t = GetWord(offset);
	    if (t == 0xffffU)
		t = 0x0fffffffUL;
	}
#if 0
	puthex(t>>16);
	puthex(t);
	puts(" c");
#endif
	frag->size += fatInfo.fatSectorsPerCluster;
	++fatCluster;
	if (frag->size >= /*32768U*/65535U-fatInfo.fatSectorsPerCluster
	    || t != fatCluster) {
	    /* if full or not sequential cluster, start next fragment */
#if 0
	    puts(" frag");
#endif
	    fatCluster = t;
	    if (frag == &fragments[MAX_FRAGMENTS-1] ||
		t == 0x0fffffffUL) {
		frag->start |= LAST_FRAGMENT;
		return frag+1;
	    }
	    frag++;
	l55:
	    frag->size = 0;
	    frag->start =
		fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

	} else {
	    offset += (fatInfo.IS_FAT_32 ? 4 : 2);
	    offset &= 511;
	    if (offset != 0)
		continue;
	}
read:
	/* If FAT page is not already in memory, load it. */
	if (fatInfo.IS_FAT_32) {
	    offset = ((u_int16)fatCluster & 0x7F) << 2;
	    ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
	} else {
	    offset = ((u_int16)fatCluster & 0xff) << 1;
	    ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
	}
    }
}
#endif

u_int16 gFileNum[2];

#ifdef __VSDSP__
auto s_int16 OpenFile(register __c0 u_int16 fileNum);
#else
auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
    register __i0 __y struct FRAGMENT * __y nextFragment;

    /* Start at the start of root directory. */

    if (fatInfo.IS_FAT_32) {
	nextFragment = FragmentList(curFragment, 2);
    } else {
	curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
	curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
	nextFragment = curFragment + 1;
    }
    gFileNum[0] = fileNum;
    gFileNum[1] = 0;
    return HandleDir(curFragment, nextFragment);
}
#endif

#ifdef CHECKFILETYPE
#if 0
#define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
auto s_int16 CheckFileType(register __a u_int32 name) {
    name &= 0x00ffffffUL;
    if (name == MKID('W','A','V')) {
	return -1;
    }
    if (name == MKID('M','I','D')) {
	return -1;
    }
    return 0;
}
#endif/*1*/
#endif/*CHECKFILETYPE*/

/*147 words*/
auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
    register __c1 int i = 0;
    register u_int32 currentSector = curFragment->start /*& 0x7fffffffUL */;

#ifdef ENABLE_WATCHDOG
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    goto first;

    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */

	/* We are now looking at FAT directory structure. */
	/* Is current file a regular file? */
	if (fn == 0) {
	    goto eod; /*end of directory*/
	}
	if (fn != 0xe5) { /* Not deleted */
#ifdef WITH_SUBDIRECTORIES
	    /* Is it a subdirectory? */
	    if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
		/* TODO: do not call if FAT12! */
		if (HandleDir(nextFragment,
			      FragmentList(nextFragment,
					   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
		    return -1;
		ReadDiskSector(currentSector);/*reread sector*/
	    } else /* it was a subdirectory */
#endif
	    /* Attributes: NO directory, NO volume id, NO system, NO hidden */
	    if ((GetByte(i+11/*Attr*/) & 0xde) == 0
#ifdef CHECKFILETYPE
		/*
		  check supported files from name suffix:
		  vs1103:  wav mid
		*/
		&& CHECKFILETYPE(GetLong(i+8))
#endif
		) {

		/* It is a regular file. */
#if defined(SCI_UI)
		if (USEX(SCI_AICTRL3) & CTRL3_BY_NAME) {
		    register int j, t = 0;
		    register u_int16 *p = stream_buffer + i/2;
		    register __y u_int16 *d = fileName;
		    p[5] &= 0xff00; /* clear the attribute byte */
		    for (j=0; j<6; j++) {
			t |= *d++ ^ *p++;
		    }
		    if (t == 0) {
			USEX(SCI_AICTRL0) = gFileNum[1];
			goto thisfile;
		    }
		}
#endif

#if 0
		gFileNum[1]++;
#else
		{ /* force gFileNum[1]++ to a different register */
		    register u_int16 t = gFileNum[1] + 1;
		    gFileNum[1] = t;
		}
#endif

		if (gFileNum[0]-- == 0) {
		thisfile:
#ifdef SCI_UI
		    /* Put the 8.3-character filename to memory so the user
		       can read it. Two bytes per word, big-endian. */
		    {   /* 13 words */
			register int j;
			register u_int16 *p = stream_buffer + i/2;
			register __y u_int16 *d = fileName;
			for (j=0;j<6;j++) {
			    *d++ = *p++;
			}
		    }
#endif
		    /* ------------ FILE FOUND ------------- */
		    fatInfo.fileSize = GetLong(i+28); /*do first!*/
		    FragmentList(&fragments[0],
				 ((u_int32)GetWord(i+20)<<16) + GetWord(i+26));
		    return -1; /* File found, All OK return */
		}
	    } /* normal file */
	} /* 0xe5 / 0x00 deleted or unused entry */
	i = (i + 32) & 511;
	if (i == 0) {
	    currentSector++;
	    if (--(curFragment->size) == 0) {
		if ((s_int16)(curFragment->start>>16) < 0) {
	eod:
#if defined(SCI_UI)
		    USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME;
#endif
		    return gFileNum[1];
		}
		curFragment++;
		currentSector = curFragment->start /*& 0x7fffffffUL*/;
	    }
	first:
#if 0
	    puthex(currentSector);
	    puts(" read");
#endif
	    ReadDiskSector(currentSector);
	}
    }
}


#ifdef HDATCHECK
u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;
u_int16 __x * __x stream_save;

#ifndef SCI_UI
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
#ifdef KEYTIMER
    s_int16 KeyTimer;
#endif
} ui;
#endif

#if 1

#ifdef SCI_UI
s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);
s_int16 ApplAddrRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);
#endif/*SCI_UI*/

#else/*1*/

#ifdef SCI_UI
s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n) {
    /* Read memory through SCI */
    if (mode == APPL_R2 /*read AICTRL2*/) {
	register s_int16 a = siAddr++;
	if (a - 0x4000 < 0)
	    USEX(SCI_AICTRL2) = USEX(a);
	else
	    USEX(SCI_AICTRL2) = USEY(a - 0x4000);
    }
    return n;
}
#endif/*SCI_UI*/
#endif

#if 1
auto u_int32 NextSector(void);
#else
auto u_int32 NextSector(void) {
    spix.sector++;
    if (--spix.fragSize == 0) {
	if ((s_int32)spix.sector < 0) /* was last fragment */
	    return -1;
	spix.fragment++;
	spix.sector   = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size; /* *fatInfo.sectorsPerCluster*/
    }
    return 0;
}
#endif


u_int16 ClockSdi(void); /* dummy return value */
#if 0
u_int16 ClockSdi(void) {
    register int i;
    for (i=0;i<32;i++) {
	USEX(SER_DREQ) = 1;
	USEX(SER_DREQ) = 0;
    }
    return -1;
}
#endif

s_int16 hcShift;
#ifdef UART_FIFO
u_int16 uiNum = 0;
s_int16 uiLoops = -1;
#endif/*UART_FIFO*/

/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

//USEX(UART_DATA) = spiState;

#if defined(UART_UI)
    if ((USEX(SCI_AICTRL3) & CTRL3_UPDATE_VOL)) {
	USEX(SCI_AICTRL3) &= ~CTRL3_UPDATE_VOL;
	SetVolumeFromSpi();
    }
#endif/*UART_UI*/

#if 0 //defined(UART_FIFO)
    /*Does not fit*/
    if (UartFill()) {
	register int cmd = UartGetByte();
#if 0
	if (cmd == 'V') { /* Set direct volume */
	    USEX(SCI_VOL) = (uiNum & 255) * 0x0101;
	    uiNum = 0;
	    goto showvol;
	} else
#endif
	if (cmd == 'P') { /* play file num */
	    USEX(SCI_AICTRL0) = uiNum;
	    uiNum = 0;
	    /*Note: \n is not received, because we restart the software.
	            So, we clear uiNum here. */
	    goto cancelpauseoff;
	} else if (cmd == '\n' || cmd == '\r') {
	    uiNum = 0;
	} else if (cmd >= '0' && cmd <= '9') {
	    uiNum = uiNum*10 + cmd - '0';
	} else if (cmd == 'f') {
	    /* file mode */
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_BEFORE_PLAY;
//	    putch('f');
	} else if (cmd == 'c') {
	    /* continuous play mode */
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
//	    putch('c');
	} else if (cmd == 'l') {
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    USEX(SCI_AICTRL3) |= CTRL3_LOOP_SONG;
	    if (uiNum) {
		uiLoops = uiNum; //loop this many times
	    }
	    uiNum = 0;
//	    putch('l');
	} else if (cmd == 'C') { /* cancel play: replay the same file */
//	    putch('C');
	    goto cancelpauseoff;
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
#if 0
	} else if (cmd == '+') {
	    if (USEX(SCI_VOL)) {
		USEX(SCI_VOL) -= 0x0101; //KeyEventHandler(ke_volumeUp2);
	    }
	    goto showvol;//putword(player.volume);
#endif
#if 0
	} else if (cmd == '-') {
	    //KeyEventHandler(ke_volumeDown2);
	    USEX(SCI_VOL) += 0x0101;
	showvol:
	    SetVolumeFromSpi();
	    putword(USEX(SCI_VOL));
#endif
	} else if (cmd == '=') {
	    /* pause */
#ifdef SCI_UI
	    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
#else
	    /* play pause/continue -- not very good for midi.. */
	    if (cachedSpiState != spi_pause) {
		oldSpiState = cachedSpiState;
		cachedSpiState = spi_pause;
		stream_save = stream_wr_pointer;
		stream_wr_pointer = stream_rd_pointer+1;
	    }
#endif/*else SCI_UI*/
	    //player.pauseOn = 1;
	    goto echo;//putch('=');
	} else if (cmd == '>') {
	    /* play */
#ifdef SCI_UI
	    USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
#else
	    if (cachedSpiState == spi_pause) {
		cachedSpiState = oldSpiState;
		/* restore 'stolen' data */
		stream_wr_pointer = stream_save;
	    }
#endif
	    //player.pauseOn = 0;
	    parametric_x.midiConfig = 0; /* play speed 0 and reverb auto */
	    goto echo;//putch('>');
#if 0
	} else if (cmd == '') {
	    /* Fast forward */
	    //if (cs.fastForward < 4)
	    //++cs.fastForward;
	    /* increase play speed */
	    if ((parametric_x.midiConfig & 0x70) < 0x70)
		parametric_x.midiConfig += 0x10;
	    putch(((s_int16)parametric_x.midiConfig>>4) & 7);
#endif
	} else if (cmd == 'n') {
	    /* next */
	    //player.nextFile = player.currentFile + 1;
	    USEX(SCI_AICTRL0)++;
	    putch('n');
#if 0
	    goto cancelpauseoff;
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
	} else if (cmd == 'p') {
	    /* previous */
	    //player.nextFile = player.currentFile - 1;
	    if (USEX(SCI_DECODE_TIME) < 5) {
		if (USEX(SCI_AICTRL0)) {
		    USEX(SCI_AICTRL0)--;
		} else {
		    USEX(SCI_AICTRL0) = USEX(SCI_AICTRL2) - 1;
		}
	    }
	    putch('p');
#endif
	cancelpauseoff:
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
#ifdef SCI_UI
	    USEX(SCI_AICTRL0) |= 0x8000U;
#else
	    ;
#endif
#if 0
	} else if (cmd == '?') {
	    /* position + other info -- Note: will block. */
	    putword(USEX(SCI_DECODE_TIME));
	    putch((fatInfo.fileSize-2*spix.size) / ((fatInfo.fileSize >> 8)+1));
#endif
	} else {
	    /*unknown command*/
	echo:
	    putch(cmd);
	}
    }
#endif/*UART_FIFO*/


#ifdef SCI_UI
    /* play pause/continue */
    if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
	if (cachedSpiState != spi_pause) {
	    oldSpiState = cachedSpiState;
#if defined(UART_UI) || defined(UART_FIFO)
	    putch('p');//USEX(UART_DATA) = 'p';
#endif
	    cachedSpiState = spi_pause;
	    stream_save = stream_wr_pointer;
	    stream_wr_pointer = stream_rd_pointer+1;
	    /* Does not matter if it is one off..
	       StreamDiff() returns the right value regardless. */
	}
    } else {
	if (cachedSpiState == spi_pause) {
#if defined(UART_UI) || defined(UART_FIFO)
	    putch('c');//USEX(UART_DATA) = 'c';
#endif
	    cachedSpiState = oldSpiState;
	    stream_wr_pointer = stream_save;
	}
    }
    if ((USEX(SCI_AICTRL0) & 0x8000U)) {
	/* new song */
	USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
	/*NEW: clear pause and file ready bits automatically */
	USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON | CTRL3_FILE_READY | CTRL3_AT_END);
	SpiSendClocks();
#if defined(UART_UI) || defined(UART_FIFO)
	putch('C');//USEX(UART_DATA) = 'C';
#endif

	/* Discard any unread data, including CRC, before starting again */
	{
	    register int i;
	    for (i=0; i<256+2; i++)
		SpiSendReceiveMmc(-1/*0xffff*/,16);
	}
	/* Jump to new track immediately */
	MyReset();
    }
#endif/*SCI_UI*/


#ifndef USE_DEBUG

#ifdef PLAYTIME
    if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
	USEX(SCI_AICTRL0)++;
#if !defined(NO_MIDI) && !defined(HDATCHECK)
	USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
#endif
	cachedSpiState = spi_zeros;
    }
#endif/*PLAYTIME*/


    if (cachedSpiState == spi_file) {
	/* extra byte read if odd file size */
	spix.size = (fatInfo.fileSize+1)/2; /* in words! */
	cachedSpiState = spi_seek;
setfrag:
	spix.fragment = &fragments[0];
	spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
	spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
    }

    if (cachedSpiState == spi_seek) {
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << hcShift);
	cachedSpiState = spi_waitdata;
	spiCnt = 0;
#ifdef HDATCHECK
	if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
	    /* if file not detected at HDATCHECK/2 kB, skip the rest */
	    goto end_of_file;
	}
#endif
    }
    if (cachedSpiState == spi_waitdata) {
	register u_int16 res = SpiSendReceiveMmc(0xff00,8);
	int k;
	if (res == 0xfe) {
	    cachedSpiState = spi_data;
	    spiCnt = 0;
	    NextSector();
	} else if (res == 0xff) {
	    if ((s_int16)++spiCnt < 0) {
		MyReset(); /* timeout, reset immediately */
	    }
	} else {
	    /* Read Error, data not transferred */
//	    putch(res);
	    goto end_of_file;
	}

    } else if (cachedSpiState == spi_data) {
	while (StreamDiff() < STREAM_BUFFER_SZ-40 &&
	       spix.size > 0) {
	    register int i, j = 32;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    if (j > spix.size)
		j = spix.size;
	    for (i=0;i<j;i++) {
		*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		cachedSpiState = spi_seek;
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
		break;
	    }
	    /* watchdog clearing here doesn't work with midi */
	}

#ifdef ENABLE_WATCHDOG
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	if (spix.size <= 0) {
end_of_file:
	    /* 8 words */
	    /* Discard any unread data, including CRC, before starting again */
	    {
		register int i;
		for (i=0; i<256+2; i++)
		    SpiSendReceiveMmc(-1/*0xffff*/,16);
	    }
	    SpiSendClocks();

#ifdef SCI_UI
	    {
		register int m = (USEX(SCI_AICTRL3)& CTRL3_PLAY_MODE_MASK);
		if (m == CTRL3_LOOP_SONG) {
		    /* restart the same file immediately */
		    cachedSpiState = spi_file;
#if defined(UART_UI) || defined(UART_FIFO)
		    putch('l');//USEX(UART_DATA) = 'l';
#endif
		} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
		    /* pause after play */
		    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* normal play: next file */
		    cachedSpiState = spi_zeros;
		}
	    }
#else/*SCI_UI*/
	    {
		register int m = (USEX(SCI_AICTRL3)& CTRL3_PLAY_MODE_MASK);
		if (m == CTRL3_LOOP_SONG) {
		    /* restart the same file immediately */
		    cachedSpiState = spi_file;
		} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
		    /* goto pause mode, so next and prev will work to
		       jump to next/prev song. */
		    oldSpiState = spi_zeros;
		    cachedSpiState = spi_pause;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* normal play: next file */
		    cachedSpiState = spi_zeros;
		}
	    }
#endif/*elseSCI_UI*/
	}
    } else if (cachedSpiState == spi_zeros) {
	if (StreamDiff() < STREAM_BUFFER_SZ-40) {
	    register int i, z = 0;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    for (i=0;i<32;i++) {
		*wr = z;
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
		    wr = stream_buffer;
		} else {
		    wr++;
		}
		++spiCnt;
		stream_wr_pointer = (void *)wr;
	    }
	    /* watchdog clearing here doesn't work with midi */
	}
	if (spiCnt >= 256+2048) { /* was left to max 256 by spi_data */
	    /* 2048 words instead of bytes */
goreset:
	    /* 8 words */
	    /* Discard any unread data, including CRC, before starting again */
	    {
		register int i;
		for (i=0; i<256+2; i++)
		    SpiSendReceiveMmc(-1/*0xffff*/,16);
	    }
	    SpiSendClocks();

#if defined(UART_UI) || defined(UART_FIFO)
	    putch('e');//USEX(UART_DATA) = 'e';
#endif
	    MyReset();
	}
#ifdef ENABLE_WATCHDOG
	goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    } else {
#ifdef ENABLE_WATCHDOG
wdogack:
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    }


#ifdef SCI_UI
    USEX(SER_DREQ) = 0;
#else

#ifndef NO_DREQ_LED
    /* 15 words */
    {
	register int t = 0;

	if ((ui.Timer & 0x18) == 0) {
	    /* DREQ indicates random play status */
#ifdef NO_RANDOM
	    t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#else
	    t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
#ifdef LOUDNESS_ALWAYS_ON_OFF
	    /* indicate random play */
#else /*LOUDNESS_ALWAYS_ON_OFF*/

	    if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
		==
#else
		!=
#endif
		spi_pause)
		t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
#endif /*else NO_RANDOM*/
	}
	USEX(SER_DREQ) = t;
    }
#endif/*!NO_DREQ_LED*/
#endif/*!SCI_UI*/
#endif/*!USE_DEBUG*/

    spiState = cachedSpiState;
#ifdef KEYTIMER
#ifdef BATTCHECK
    {
	register s_int16 batt = USEX(MODU_DATA);
	if (ui.KeyTimer >= 0 ||
	    ((ui.KeyTimer & 8) == 0 &&
	     (batt < 0x240 || ui.KeyTimer > (batt>>7) - 60)))
	    USEX(UART_DATA) = 0x00;
    }
#else
    if (ui.KeyTimer >= 0) {
	USEX(UART_DATA) = 0x00;
    }
#endif
#endif/*KEYTIMER*/
}


auto void MyMain(register s_int16 __a1 fullReset) {
#ifdef BATTCHECK
    USEX(MODU_DIV) = 0x0030;
#endif/*BATTCHECK*/

#if defined(SCI_UI) && defined(ENABLE_WATCHDOG)
    USEX(WDOG_CONFIG) = 1000;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif

#if defined(SCI_UI)
    USEX(SCI_AICTRL2) = 0;
#endif
#if 0 && defined(VS1103)
    USEX(SCI_MODE) = (1<<SCIMB_SDI_NEWMODE)
	+ (0<<SCIMB_CONF); /*0=SDIMIDI,SCIWAV 1=SDIWAV,SCIMIDI*/
#endif

    USEX(SCI_MODE) &= ~(1<<SCIMB_OUT_OF_WAV);
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2|SPI_CLK; /*use same= -1 word*/

    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2;


#ifdef NO_RANDOM
    InitHardware(); /* clears memory, so can't be used for debugging */
#else
    {
	register int r = mrand_seed;
#if defined(KEYTIMER)
	register int t = ui.KeyTimer;
#endif
	InitHardware(); /* clears memory, so can't be used for debugging */
#ifdef KEYTIMER
	if (fullReset)
	    t = 0;
	ui.KeyTimer = t;
	USEX(UART_DIV) = 0xffff;//0x3f46; /* 9600 */
#endif
	mrand_seed = r;
    }
#endif/*else NO_RANDOM*/

#if defined(UART_UI) || defined(UART_FIFO)
//    putch('R');
#endif

#if 1 /*VERSION*/
    /*
      v0.6 print version string
    */
    putword(('0'<<8)|'.');
    putword(('6'<<8)|'0');
    putch('\n');
#endif
#ifdef UART_FIFO
    UartInit();
#endif
#ifdef UART_UI
    fullReset = MyRxInt; /* Pull in replacement UART RX interrupt handler. */
#endif


#if 0
    uartByteSpeed = 960;
    SetClockFromSpi();
    putch('s');
#endif

    InitCommon();
    //AudioInit(); //rather use..
    //AudioSet(2, 44100U);

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
//    USEX(INT_ENABLE) = 0;
//    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
//    USEX(INT_GLOB_ENA) = 0;

    AudioSet(2, 44100U);
    InitReadFrame();

    /* wake up MMC! -- If this takes too long, watchdog will bite. */
    /* inlining InitStorage() saves 18 words */

    /*
      Kingston 128M (Korea) -- supports interrupted read
      Cmd(7B00000000)
      00FF               (timeout waiting for reply)
      Cmd(4000000000)
      000D               (crc error, invalid command)
      not valid idle
      Cmd(7B00000000)
      0005               (crc error)
      Cmd(4000000000)
      0001               (ok)
    */

    //while (1)
    {
	register s_int16 i, cmd;
	//USEX(UART_DATA) = 'I';
    tryagain:
        for (i=0; i<512; i++) {
	    SpiSendClocks();
	}
	/* Why, oh why doesn't this work with some uSD cards when the same
	   code works with VS1000 Audio Module? MmcCommand seems to be a
	   virtual carbon-copy between the versions.

	   It seems some capacitance on SDCLK helps with some cards.
	   It certainly isn't due to data/clock timing. It seems to be
	   that the uSD card detects extra clock edges, because sometimes
	   when right "on the edge" the card responds
	*/
	/* MMC Init, command 0x40 should return 0x01 if all is ok. */
	i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
	//putch(i);
	if (i != 1) {
	    BusyWait10();
//USEX(UART_DATA) = i;
	    goto tryagain;//continue; /* No valid idle response */
	}
	//USEX(UART_DATA) = '1';
	cmd = MMC_SEND_OP_COND|0x40;
	//MmcCommand(MMC_CRC_ON_OFF/*CMD59*/|0x40, 0);
	if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	    == 1) {
	    /* MMC answers: 0x05 SD answers: 0x01 */
	    /* Should we read the whole R7 response? */
	    //SpiSendReceiveMmc(-1, 32);
	    cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
	}
#if 0
	/* Some cards require CMD58 after CMD8 -- does this still work with MMC?*/
	MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0);
	SpiSendReceiveMmc(-1, 16);
	SpiSendReceiveMmc(-1, 16);
#endif
	/* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
	//i = 0; /* i is 1 when entered .. but does not make the code shorter*/
	while (1) {
	    register int c;
	    if (cmd == (0x40|41)) {
		MmcCommand(0x40|55/*CMD55*/,0);
	    }
	    c = MmcCommand(cmd/*SD_SEND_OP_COND|0x40*/, 0x40100000UL);
	    if (c == 0) {
		break;
	    }
	    if (++i < 0 /*>= 32000*//*12000*//*< 0*/ || c != 1) {
	      BusyWait10();
	      //USEX(UART_DATA) = '2';
	      goto tryagain; /* Not able to power up mmc */
	    }
	}

	hcShift = 9;
	i = 0x0246;
	/* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
	/* Support HC for SD, but not for MMC! */
	if (cmd == (0x40|41) &&
	    MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
	    if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
		/* OCR[30]:CCS - card capacity select */
		/* HighCapacity! */
		/* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
		i = 6;
		hcShift = 0;
	    }
	    //SpiSendReceiveMmc(-1, 16);
	}

#if 0
	/* Set Block Size of 512 bytes -- default for at least HC */
	/* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
	/* Must check return value! (MicroSD restart!) */
	if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
	    goto tryagain;
	/* All OK return */
#endif
    }

    /* Detects cards without partition table */
    if (InitFileSystem() == 0) {
#ifdef SCI_UI
	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	    USEX(SCI_AICTRL2) = 0x7fff;
	} else
#endif/*SCI_UI*/
	{
	    /* determine number of playable files */
	    USEX(SCI_AICTRL2) = OpenFile(0xffffU);
	}

#if !defined(NO_RANDOM) /* no random play feature */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
	    /* random play (not shuffle!) */
	    USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL2));
	}
#endif
	if (OpenFile(USEX(SCI_AICTRL0)
#ifdef SCI_UI
		      &= 0x7fffU //max 32768 files in SCI-controlled version
#endif
	    ) < 0) {

#if defined(SCI_UI)
	    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		CTRL3_PAUSE_BEFORE_PLAY) {
		/* Mark file ready and go to pause state in UserHook */
		USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
	    }
	    /*NEW: set FILE_READY always */
	    USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
		(USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
#endif


#if defined(PAUSE_AT_POWERON) /*pause at power-on, then use pause after play*/
	    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		CTRL3_PAUSE_BEFORE_PLAY) {
		USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
		oldSpiState = spi_file;
		spiState = spi_pause;
		stream_save = stream_wr_pointer;
	    } else {
		spiState = spi_file;
	    }
#else /*PAUSE_AT_POWERON*/
	    spiState = spi_file;
#endif /*else PAUSE_AT_POWERON*/

#ifdef HDATCHECK
	    hdatCheck = 0;
#endif
	} else {
	gofirst:
	    USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
	    MyReset();
	}
    } else {
	/* Initialize again */
	goto gofirst;//MyReset();
    }

#if !defined(SCI_UI) && defined(WAIT_FOR_BUTTON)
    if ((USEX(SCI_AICTRL3)&CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
	/* Only once per boot */
	USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	while (1) {
	    ui.Data = 0;
	    /* toggle MOSI to clock SDI */
	    USEX(INT_ENABLE) = ENABLE_SCI_SDI_DAC_RX;
	    ClockSdi();
	    if (ui.Data == 0) {
		USEX(GPIO_ODATA) &= ~SPI_xCS;
		ClockSdi();
		USEX(GPIO_ODATA) |= SPI_xCS;
		ui.Data >>= 1;/* 0xffff or 0x8000 ? -> 0x7fff 0x4000 */
	    }
	    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
#ifdef ENABLE_WATCHDOG
	    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	    if (ui.Data) {
		ui.Key = ui.Data;
		ui.KeyTime = -32767;
		break;
	    }
	    //Halt();//Sleep();
	}
    }
#endif

#if !defined(SCI_UI) && defined(WAIT_FOR_UART_BYTE)
    /*TODO: lower clock */
    {
	register u_int16 old = USEX(SCI_CLOCKF);
	USEX(SCI_CLOCKF) = 0;
	SetClockFromSpi();

	/*rescue rx if framing error happened.*/
	USEX(UART_DATA);
	USEX(INT_ENABLE) = 0;
	if (1) {
	    register u_int16 d = 0x7f7f;//USEX(UART_DIV);
	    register int i;
	    for (i=0;i<256;i++)
		USEX(UART_DIV) = d;
	}
	USEX(SER_DREQ) = 0;
	while (1) {
	    USEX(SER_DREQ) ^= 1;
#ifdef ENABLE_WATCHDOG
	    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	    if (USEX(UART_STATUS) & UART_ST_RXFULL) {
		USEX(UART_DATA);
		break;
	    }
	}
	USEX(SCI_CLOCKF) = old;
	SetClockFromSpi();
    }
#endif


#if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
    /*otherwise keep it in powerdown until MMC found*/
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
#endif


    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;

    /*A modified tradpatch for VS1103B */
    {
#define MIXER_OUT_SAMPLES 34
	auto s_int16 DecodeHeader(register __d u_int32 head);
	struct Wav {
	    s_int16 channels;
	    u_int16 sampleRate;
	    s_int16 ima;
	    s_int16 bitsPerSample;
	    s_int16 samplesPerBlock;
	    u_int16 imaSync;
	    u_int32 samplesLeft;
	    u_int16 flags;
	};
	extern struct Wav wav;
	extern s_int16 mixerOut[3][2*MIXER_OUT_SAMPLES];

	register u_int32 head = 0;
	skipMixer = 1;
	while (1) {
	    if (StreamDiff()) {
		head <<= 8;
		head |= SdiGetC();
		if (head == MAKE_ID('R','I','F','F')) {
		    ResetAdpcm();
#if 0
		    /* Move "RIFF" back to input buffer */
		    if ((stream_rd_pointer -= 2) < stream_buffer)
			stream_rd_pointer += STREAM_BUFFER_SZ;
#else
		    /* we know we always start from the beginning */
		    stream_rd_pointer = stream_buffer;
#endif
		    adpcmSource = 0;	/* ADPCM from SDI */
		    while (1) {
			GetAdpcmSamples(NULL, mixerOut[0], 32);
			if (wav.sampleRate && wav.sampleRate != hwSampleRate) {
			    SetHardware(wav.channels, wav.sampleRate);
			}
			WmaStereoCopy(mixerOut[0], 32);
		    }
		} else {
		    DecodeHeader(head);
		}
	    } else {
#if 1
		Sleep(); //Audio buffer is empty at start, so can just Sleep();
#else
		memset(mixerOut[1], 0, 64);
		WmaStereoCopy(mixerOut[1], 32);
#endif
	    }
	}
    }
}
