/*
  VS1103b Simple SPI Flash Recorder / Player
  ==========================================
  This source code package is provided as-is and without warranty.
  (C)2008 VLSI Solution Oy

  This code is licensed for free use with any product using VS1103b.

  Some features can be turned on and off with preprocessor definitions
  in Makefile and in this file. Features that are enabled have an
  effect on the code size, and the resulting program may not fit in
  the available code space. In this case the linker will give an error.
*/

/*
  Note on MOSI/DREQ:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  In this version SDI interrupt is not enabled, so it does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "module.h"
#include "h1103/ima.h"
#include "h1103/imalib.h"
#include "spi.h"
#undef puts
#undef printf
#undef putchar
#undef fputs


auto u_int16 SpiReadWord(u_int32 byteAddr) {
  u_int16 res;

  while (SpiPrivStatus() & 1)
    SpiPrivDelay();

  USEX(GPIO_ODATA) |= SPI_xCS;
  SpiPrivDelay();
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  SpiPrivSendReceive(SPI_READ, 8);
  SpiPrivDelay();
  SpiPrivSendReceive(byteAddr >> 8, 16);	// 24-bit address bits 23:8
  SpiPrivSendReceive(byteAddr & 0xFF, 8);	// 24-bit address bits  7:0
  SpiPrivDelay();
  res = SpiPrivSendReceive(0, 16);
  SpiPrivDelay();
  USEX(GPIO_ODATA) |= SPI_xCS;

  return res;
}
